"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.forwardWebSocketRequest = void 0;
const token_1 = require("../utils/token");
const logger_1 = require("../../logs/logger");
const downstream_post_stream_to_server_1 = require("../http/downstream-post-stream-to-server");
const metrics_1 = require("../utils/metrics");
const version_1 = __importDefault(require("../utils/version"));
const scm_1 = require("../../client/scm");
const log_1 = require("../../logs/log");
const json_1 = require("../utils/json");
const replace_vars_1 = require("../utils/replace-vars");
const url_1 = require("url");
const url_2 = require("url");
const try_json_parse_1 = __importDefault(require("../utils/try-json-parse"));
const undefsafe_1 = __importDefault(require("undefsafe"));
const request_1 = require("../http/request");
const filtersAsync_1 = require("../filter/filtersAsync");
const forwardWebSocketRequest = (options, io) => {
    // 1. Request coming in over websocket conn (logged)
    // 2. Filter for rule match (log and block if no match)
    // 3. Relay over HTTP conn (logged)
    // 4. Get response over HTTP conn (logged)
    // 5. Send response over websocket conn
    var _a;
    //   const filters = Filters(options.filters?.private);
    const filters = (0, filtersAsync_1.loadFilters)((_a = options.filters) === null || _a === void 0 ? void 0 : _a.private);
    return (brokerToken) => async (payload, emit) => {
        var _a, _b, _c;
        const requestId = payload.headers['snyk-request-id'];
        const logContext = {
            url: payload.url,
            requestMethod: payload.method,
            requestHeaders: payload.headers,
            requestId,
            streamingID: payload.streamingID,
            maskedToken: (0, token_1.maskToken)(brokerToken),
            hashedToken: (0, token_1.hashToken)(brokerToken),
            transport: (_c = (_b = (_a = io === null || io === void 0 ? void 0 : io.socket) === null || _a === void 0 ? void 0 : _a.transport) === null || _b === void 0 ? void 0 : _b.name) !== null && _c !== void 0 ? _c : 'unknown',
        };
        if (!requestId) {
            // This should be a warning but older clients won't send one
            // TODO make this a warning when significant majority of clients are on latest version
            logger_1.log.trace(logContext, 'Header Snyk-Request-Id not included in headers passed through');
        }
        const websocketEmit = emit;
        const overrideEmit = (responseData, isResponseFromRequestModule = false) => {
            var _a;
            // This only works client -> server, it's not possible to post data server -> client
            logContext.requestMethod = '';
            logContext.requestHeaders = {};
            if ((_a = io === null || io === void 0 ? void 0 : io.capabilities) === null || _a === void 0 ? void 0 : _a.includes('receive-post-streams')) {
                // Traffic over HTTP Post
                const postHandler = new downstream_post_stream_to_server_1.BrokerServerPostResponseHandler(logContext, options.config, brokerToken, payload.streamingID, options.config.serverId, requestId);
                if (isResponseFromRequestModule) {
                    logger_1.log.trace(logContext, 'posting streaming response back to Broker Server');
                    postHandler.forwardRequest(responseData);
                }
                else {
                    // Only for responses generated internally in the Broker Client/Server
                    postHandler.sendData(responseData);
                }
            }
            else {
                // Traffic over websockets
                if (payload.streamingID) {
                    if (isResponseFromRequestModule) {
                        legacyStreaming(logContext, responseData, options.config, io, payload.streamingID);
                    }
                    else {
                        io.send('chunk', payload.streamingID, '', false, {
                            status: responseData.status,
                            headers: responseData.headers,
                        });
                        io.send('chunk', payload.streamingID, JSON.stringify(responseData.body), true);
                    }
                }
                else {
                    logger_1.log.trace({ ...logContext, responseData }, 'sending fixed response over WebSocket connection');
                    websocketEmit(responseData);
                }
            }
        };
        emit = overrideEmit;
        logger_1.log.info(logContext, 'received request over websocket connection');
        const prepareRequestFromFilterResult = async (result) => {
            var _a;
            (0, metrics_1.incrementWebSocketRequestsTotal)(false);
            if (result.url.startsWith('http') === false) {
                result.url = 'https://' + result.url;
                logContext.resultUrlSchemeAdded = true;
            }
            logContext.httpUrl = result.url;
            if (!payload.headers['user-agent']) {
                payload.headers['user-agent'] = 'Snyk Broker ' + version_1.default;
                logContext.userAgentHeaderSet = true;
            }
            if (result.auth) {
                payload.headers['authorization'] = result.auth;
                logContext.authHeaderSetByRuleAuth = true;
            }
            else {
                const parsed = (0, url_1.parse)(result.url);
                if (parsed.auth) {
                    // if URL contains basic auth,
                    // remove authorization header to prefer auth on the URL.
                    if (parsed.auth.includes(':')) {
                        delete payload.headers['authorization'];
                    }
                    // if URL contains token auth,
                    // put the token in the authorization header
                    // instead of on the URL.
                    else {
                        payload.headers['authorization'] = `token ${parsed.auth}`;
                        // then strip from the url
                        parsed.auth = null;
                        result.url = (0, url_2.format)(parsed);
                    }
                    logContext.authHeaderSetByRuleUrl = true;
                }
            }
            // if the request is all good - and at this point it is, we'll check
            // whether we want to do variable substitution on the body
            //
            // Variable substitution - for those who forgot - is substituting a part
            // of a given string (e.g. "${SOME_ENV_VAR}/rest/of/string")
            // with an env var of the same name (SOME_ENV_VAR).
            // This is used (for example) to substitute the snyk url
            // with the broker's url when defining the target for an incoming webhook.
            if (!options.config.disableBodyVarsSubstitution && payload.body) {
                const parsedBody = (0, try_json_parse_1.default)(payload.body);
                if (parsedBody.BROKER_VAR_SUB) {
                    logContext.bodyVarsSubstitution = parsedBody.BROKER_VAR_SUB;
                    for (const path of parsedBody.BROKER_VAR_SUB) {
                        let source = (0, undefsafe_1.default)(parsedBody, path); // get the value
                        source = (0, replace_vars_1.replace)(source, options.config); // replace the variables
                        (0, undefsafe_1.default)(parsedBody, path, source); // put it back in
                    }
                    payload.body = JSON.stringify(parsedBody);
                }
            }
            if (!options.config.disableHeaderVarsSubstitution &&
                payload.headers &&
                payload.headers['x-broker-var-sub']) {
                // check whether we want to do variable substitution on the headers
                logContext.headerVarsSubstitution = payload.headers['x-broker-var-sub'];
                for (const path of payload.headers['x-broker-var-sub'].split(',')) {
                    let source = (0, undefsafe_1.default)(payload.headers, path.trim()); // get the value
                    source = (0, replace_vars_1.replace)(source, options.config); // replace the variables
                    (0, undefsafe_1.default)(payload.headers, path.trim(), source); // put it back in
                }
            }
            // remove headers that we don't want to relay
            // (because they corrupt the request)
            [
                'x-forwarded-for',
                'x-forwarded-proto',
                'content-length',
                'host',
                'accept-encoding',
                'content-encoding',
            ].map((_) => delete payload.headers[_]);
            if (options.config.removeXForwardedHeaders === 'true') {
                for (const key in payload.headers) {
                    if (key.startsWith('x-forwarded-')) {
                        delete payload.headers[key];
                    }
                }
                if (payload.headers['forwarded']) {
                    delete payload.headers['forwarded'];
                }
            }
            if (brokerToken && (io === null || io === void 0 ? void 0 : io.socketType) === 'server') {
                Object.assign(payload.headers, { 'X-Broker-Token': brokerToken });
            }
            logger_1.log.debug(logContext, 'sending websocket request over HTTP connection');
            // Sometimes we receive the body as a {type, data} object
            // Unsure why - possibly Primus?
            if (((_a = payload.body) === null || _a === void 0 ? void 0 : _a.type) === 'Buffer')
                payload.body = Buffer.of(payload.body.data);
            // Request library is buggy and will throw an error if we're POST'ing an empty body without an explicit Content-Length header
            if (!payload.body || payload.body.length === 0) {
                payload.headers['Content-Length'] = '0';
            }
            payload.headers['connection'] = 'Keep-Alive';
            payload.headers['Keep-Alive'] = 'timeout=60, max=1000';
            if ((0, scm_1.gitHubTreeCheckNeeded)(options.config, {
                method: payload.method,
                url: payload.url,
            })) {
                try {
                    (0, scm_1.validateGitHubTreePayload)(payload.body);
                }
                catch (error) {
                    logger_1.log.error({ error }, 'error while checking github tree payload for symlinks');
                    return emit({
                        status: 401,
                        body: error.message,
                    });
                }
            }
            if ((0, scm_1.gitHubCommitSigningEnabled)(options.config, {
                method: payload.method,
                url: payload.url,
            })) {
                try {
                    payload.body = await (0, scm_1.signGitHubCommit)(options.config, payload.body);
                }
                catch (error) {
                    logger_1.log.error({ error }, 'error while signing github commit');
                }
            }
            const req = {
                url: result.url,
                headers: payload.headers,
                method: payload.method,
                body: payload.body,
            };
            return req;
        };
        const makePostStreamingRequest = async (req) => {
            // this is a streaming request for binary data
            logger_1.log.debug(logContext, 'serving stream request');
            try {
                const downstreamRequestHandler = (0, request_1.getRequestToDownstream)()(req);
                emit(downstreamRequestHandler, true);
            }
            catch (e) {
                logger_1.log.error({
                    ...logContext,
                    error: e,
                    stackTrace: new Error('stacktrace generator').stack,
                }, 'caught error sending HTTP response over WebSocket');
            }
            return;
        };
        const makeLegacyRequest = (req) => {
            // not main http post flow
            (0, request_1.getRequestToDownstream)()(req, (error, response, responseBody) => {
                if (error) {
                    (0, log_1.logError)(logContext, error);
                    return emit({
                        status: 500,
                        body: error.message,
                    });
                }
                const contentLength = responseBody.length;
                // Note that the other side of the request will also check the length and will also reject it if it's too large
                // Set to 20MB even though the server is 21MB because the server looks at the total data travelling through the websocket,
                // not just the size of the body, so allow 1MB for miscellaneous data (e.g., headers, Primus overhead)
                const maxLength = parseInt(options.config.socketMaxResponseLength) || 20971520;
                if (contentLength && contentLength > maxLength) {
                    const errorMessage = `body size of ${contentLength} is greater than max allowed of ${maxLength} bytes`;
                    (0, log_1.logError)(logContext, {
                        errorMessage,
                    });
                    return emit({
                        status: 502,
                        errorType: 'BODY_TOO_LARGE',
                        originalBodySize: contentLength,
                        body: {
                            message: errorMessage,
                        },
                    });
                }
                const status = (response && response.statusCode) || 500;
                if (options.config.RES_BODY_URL_SUB && (0, json_1.isJson)(response.headers)) {
                    const replaced = (0, replace_vars_1.replaceUrlPartialChunk)(responseBody, null, options.config);
                    responseBody = replaced.newChunk;
                }
                (0, log_1.logResponse)(logContext, status, response, options.config);
                emit({ status, body: responseBody, headers: response.headers });
            });
        };
        const filterResponse = filters(payload);
        if (!filterResponse) {
            (0, metrics_1.incrementWebSocketRequestsTotal)(true);
            const reason = 'Response does not match any accept rule, blocking websocket request';
            logContext.error = 'blocked';
            logger_1.log.warn(logContext, reason);
            return emit({
                status: 401,
                body: {
                    message: 'blocked',
                    reason,
                    url: payload.url,
                },
            });
        }
        else {
            const req = await prepareRequestFromFilterResult(filterResponse);
            payload.streamingID
                ? await makePostStreamingRequest(req)
                : makeLegacyRequest(req);
        }
    };
};
exports.forwardWebSocketRequest = forwardWebSocketRequest;
const legacyStreaming = (logContext, rqst, config, io, streamingID) => {
    let prevPartialChunk;
    let isResponseJson;
    logger_1.log.warn(logContext, 'server did not advertise received-post-streams capability - falling back to legacy streaming');
    // Fall back to older streaming method if somehow connected to older server version
    rqst
        .on('response', (response) => {
        const status = (response && response.statusCode) || 500;
        (0, log_1.logResponse)(logContext, status, response, config);
        isResponseJson = (0, json_1.isJson)(response.headers);
        io.send('chunk', streamingID, '', false, {
            status,
            headers: response.headers,
        });
    })
        .on('data', (chunk) => {
        if (config.RES_BODY_URL_SUB && isResponseJson) {
            const { newChunk, partial } = (0, replace_vars_1.replaceUrlPartialChunk)(Buffer.from(chunk).toString(), prevPartialChunk, config);
            prevPartialChunk = partial;
            chunk = newChunk;
        }
        io.send('chunk', streamingID, chunk, false);
    })
        .on('end', () => {
        io.send('chunk', streamingID, '', true);
    })
        .on('error', (error) => {
        (0, log_1.logError)(logContext, error);
        io.send('chunk', streamingID, error.message, true, {
            status: 500,
        });
    });
};
//# sourceMappingURL=forwardWebsocketRequest.js.map